from django.db import models
from admin_async_upload.widgets import ResumableAdminWidget
from admin_async_upload.fields import FormResumableFileField


class AsyncFileField(models.FileField):

    def formfield(self, **kwargs):
        defaults = {'form_class': FormResumableFileField}
        if self.model and self.name:
            defaults['widget'] = ResumableAdminWidget(attrs={
                'model': self.model,
                'field_name': self.name})
        kwargs.update(defaults)
        return super(AsyncFileField, self).formfield(**kwargs)



from admin_async_upload.storage import ResumableStorage
from os.path import splitext
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


class StorageFileValidator(object):
    """
    Validation of uploaded files.

    Files uploaded using the library are passed to application with their names only.
    Any validation must happen either on the client-side or requires upload to be completed
    and file saved in application storage.
    """
    messages = {
        'file': _(u"File {name} does not exist."),
        'extension': _(u"Extension {extension} not allowed. Allowed extensions are: {allowed_extensions}"),
        'min_size': _(u"File {name} too small ({size} bytes). The minimum file size is {min_size} bytes."),
        'max_size': _(u"File {name} too large ({size} bytes). The maximum file size is {max_size} bytes."),
    }

    def __init__(self, min_size=0, max_size=None, allowed_extensions=None):
        self.min_size = min_size
        self.max_size = max_size
        self.allowed_extensions = allowed_extensions or []

    def get_storage(self):
        return ResumableStorage().get_persistent_storage()

    def validate_extension(self, value):
        ext = splitext(value)[1].lower()
        if self.allowed_extensions and ext not in self.allowed_extensions:
            message = self.messages['extension'].format(**{
                'extension': ext,
                'allowed_extensions': ', '.join(self.allowed_extensions)
            })
            raise ValidationError(message)

    def validate_exists(self, value, storage):
        if not storage.exists(value):
            message = self.messages['file'].format(**{
                'name': value,
            })
            raise ValidationError(message)

    def validate_size(self, value, storage):
        size = storage.size(value)
        if size > self.max_size:
            message = self.messages['max_size'].format(**{
                'name': value,
                'size': size,
                'max_size': self.max_size,
            })
            raise ValidationError(message)
        elif size < self.min_size:
            message = self.messages['min_size'].format(**{
                'name': value,
                'size': size,
                'min_size': self.min_size,
            })
            raise ValidationError(message)

    def __call__(self, value):
        assert type(value) == unicode
        storage = self.get_storage()
        self.validate_exists(value, storage)
        self.validate_extension(value)
        self.validate_size(value, storage)



from django.core.exceptions import ValidationError
from django.forms import fields

from admin_async_upload.widgets import ResumableAdminWidget


class FormResumableFileField(fields.FileField):
    widget = ResumableAdminWidget

    def to_python(self, data):
        if self.required:
            if not data or data == "None":
                raise ValidationError(self.error_messages['empty'])
        return data



import datetime
import posixpath
from django.conf import settings
from django.core.files.storage import FileSystemStorage
from django.utils.encoding import force_str


class ResumableStorage(object):

    def __init__(self):
        self.persistent_storage_class_name = getattr(settings, 'ADMIN_RESUMABLE_STORAGE', None) or \
                                        getattr(settings, 'DEFAULT_FILE_STORAGE',
                                                'django.core.files.storage.FileSystemStorage')

        # get_storage_class removed in Django 5.1:
        # self.chunk_storage_class_name = getattr(
        #     settings,
        #     'ADMIN_RESUMABLE_CHUNK_STORAGE',
        #     'django.core.files.storage.FileSystemStorage'
        # )

    def get_chunk_storage(self, *args, **kwargs):
        """
        Returns storage class specified in settings as ADMIN_RESUMABLE_CHUNK_STORAGE.
        Defaults to django.core.files.storage.FileSystemStorage.
        Chunk storage should be highly available for the server as saved chunks must be copied by the server
        for saving merged version in persistent storage.
        """
        # get_storage_class removed in Django 5.1:
        # storage_class = get_storage_class(self.chunk_storage_class_name)
        # return storage_class(*args, **kwargs)
        return FileSystemStorage(*args, **kwargs)

    def get_persistent_storage(self, *args, **kwargs):
        """
        Returns storage class specified in settings as ADMIN_RESUMABLE_STORAGE
        or DEFAULT_FILE_STORAGE if the former is not found.

        Defaults to django.core.files.storage.FileSystemStorage.
        """
        # get_storage_class removed in Django 5.1:
        # storage_class = get_storage_class(self.persistent_storage_class_name)
        # return storage_class(*args, **kwargs)
        return FileSystemStorage(*args, **kwargs)

    def full_filename(self, filename, upload_to, instance=None):
        if callable(upload_to):
            filename = upload_to(instance, filename)
        else:
            dirname = force_str(datetime.datetime.now().strftime(force_str(upload_to)))
            filename = posixpath.join(dirname, filename)
        return self.get_persistent_storage().generate_filename(filename)



from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models.fields.files import FieldFile
from django.forms import FileInput, CheckboxInput, forms
from django.template import loader
from django.templatetags.static import static
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy

from admin_async_upload.storage import ResumableStorage


class ResumableBaseWidget(FileInput):
    template_name = 'admin_resumable/admin_file_input.html'
    clear_checkbox_label = gettext_lazy('Clear')

    def render(self, name, value, attrs=None, **kwargs):
        persistent_storage = ResumableStorage().get_persistent_storage()
        if value:
            if isinstance(value, FieldFile):
                value_name = value.name
            else:
                value_name = value
            file_name = value
            file_url = mark_safe(persistent_storage.url(value_name))

        else:
            file_name = ""
            file_url = ""

        chunk_size = getattr(settings, 'ADMIN_RESUMABLE_CHUNKSIZE', "1*1024*1024")
        show_thumb = getattr(settings, 'ADMIN_RESUMABLE_SHOW_THUMB', False)
        simultaneous_uploads = getattr(settings, 'ADMIN_SIMULTANEOUS_UPLOADS', 3)

        content_type_id = ContentType.objects.get_for_model(self.attrs['model']).id

        context = {
            'name': name,
            'value': value,
            'id': attrs['id'],
            'chunk_size': chunk_size,
            'show_thumb': show_thumb,
            'field_name': self.attrs['field_name'],
            'content_type_id': content_type_id,
            'file_url': file_url,
            'file_name': file_name,
            'simultaneous_uploads': simultaneous_uploads,
        }

        instance = self.attrs.get('instance')
        if instance and instance.pk:
            context['instance_id'] = instance.pk

        if not self.is_required:
            template_with_clear = '<span class="clearable-file-input">%(clear)s ' \
                                  '<label for="%(clear_checkbox_id)s">%(clear_checkbox_label)s</label></span>'
            substitutions = {
                'clear_checkbox_id': attrs['id'] + "-clear-id",
                'clear_checkbox_name': attrs['id'] + "-clear",
                'clear_checkbox_label': self.clear_checkbox_label
            }
            substitutions['clear'] = CheckboxInput().render(
                substitutions['clear_checkbox_name'],
                False,
                attrs={'id': substitutions['clear_checkbox_id']}
            )
            clear_checkbox = mark_safe(template_with_clear % substitutions)
            context.update({'clear_checkbox': clear_checkbox})
        return loader.render_to_string(self.template_name, context)

    def value_from_datadict(self, data, files, name):
        if not self.is_required and data.get("id_" + name + "-clear"):
            return False  # False signals to clear any existing value, as opposed to just None
        if data.get(name, None) in ['None', 'False']:
            return None
        return data.get(name, None)


class ResumableAdminWidget(ResumableBaseWidget):
    @property
    def media(self):
        js = ["resumable.js"]
        return forms.Media(js=[static("admin_resumable/js/%s" % path) for path in js])


class ResumableWidget(ResumableBaseWidget):
    template_name = 'admin_resumable/user_file_input.html'



# -*- coding: utf-8 -*-
import fnmatch
import tempfile

from django.contrib.contenttypes.models import ContentType
from django.core.files import File
from django.utils.functional import cached_property

from admin_async_upload.storage import ResumableStorage


class ResumableFile(object):
    """
    Handles file saving and processing.
    It must only have access to chunk storage where it saves file chunks.
    When all chunks are uploaded it collects and merges them returning temporary file pointer
    that can be used to save the complete file to persistent storage.

    Chunk storage should preferably be some local storage to avoid traffic
    as files usually must be downloaded to server as chunks and re-uploaded as complete files.
    """

    def __init__(self, field, user, params):
        self.field = field
        self.user = user
        self.params = params
        self.chunk_suffix = "_part_"

    @cached_property
    def resumable_storage(self):
        return ResumableStorage()

    @cached_property
    def persistent_storage(self):
        return self.resumable_storage.get_persistent_storage()

    @cached_property
    def chunk_storage(self):
        return ResumableStorage().get_chunk_storage()

    @property
    def storage_filename(self):
        instance_id = self.params.get("instance_id")
        if instance_id:
            instance = self.field.model.objects.filter(pk=instance_id).first()
        else:
            instance = None
        return self.resumable_storage.full_filename(self.filename, self.upload_to, instance=instance)

    @property
    def upload_to(self):
        return self.field.upload_to

    @property
    def chunk_exists(self):
        """
        Checks if the requested chunk exists.
        """
        return self.chunk_storage.exists(self.current_chunk_name) and \
               self.chunk_storage.size(self.current_chunk_name) == int(self.params.get('resumableCurrentChunkSize'))

    @property
    def chunk_names(self):
        """
        Iterates over all stored chunks.
        """
        chunks = []
        files = sorted(self.chunk_storage.listdir('')[1])
        for file in files:
            if fnmatch.fnmatch(file, '%s%s*' % (self.filename,
                                                self.chunk_suffix)):
                chunks.append(file)
        return chunks

    @property
    def current_chunk_name(self):
        # TODO: add user identifier to chunk name
        return "%s%s%s" % (
            self.filename,
            self.chunk_suffix,
            self.params.get('resumableChunkNumber').zfill(4)
        )

    def chunks(self):
        """
        Iterates over all stored chunks.
        """
        # TODO: add user identifier to chunk name
        files = sorted(self.chunk_storage.listdir('')[1])
        for file in files:
            if fnmatch.fnmatch(file, '%s%s*' % (self.filename,
                                                self.chunk_suffix)):
                yield self.chunk_storage.open(file, 'rb').read()

    def delete_chunks(self):
        [self.chunk_storage.delete(chunk) for chunk in self.chunk_names]

    @property
    def file(self):
        """
        Merges file and returns its file pointer.
        """
        if not self.is_complete:
            raise Exception('Chunk(s) still missing')
        outfile = tempfile.NamedTemporaryFile("w+b")
        for chunk in self.chunk_names:
            outfile.write(self.chunk_storage.open(chunk).read())
        return outfile

    @property
    def filename(self):
        """
        Gets the filename.
        """
        # TODO: add user identifier to chunk name
        filename = self.params.get('resumableFilename')
        if '/' in filename:
            raise Exception('Invalid filename')
        value = "%s_%s" % (self.params.get('resumableTotalSize'), filename)
        return value

    @property
    def is_complete(self):
        """
        Checks if all chunks are already stored.
        """
        return int(self.params.get('resumableTotalSize')) == self.size

    def process_chunk(self, file):
        """
        Saves chunk to chunk storage.
        """
        if self.chunk_storage.exists(self.current_chunk_name):
            self.chunk_storage.delete(self.current_chunk_name)
        self.chunk_storage.save(self.current_chunk_name, file)

    @property
    def size(self):
        """
        Gets size of all chunks combined.
        """
        size = 0
        for chunk in self.chunk_names:
            size += self.chunk_storage.size(chunk)
        return size

    def collect(self):
        actual_filename = self.persistent_storage.save(self.storage_filename, File(self.file))
        self.delete_chunks()
        return actual_filename



from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponse
from django.utils.functional import cached_property
from django.views.generic import View
from admin_async_upload.files import ResumableFile


class UploadView(View):
    # inspired by another fork https://github.com/fdemmer/django-admin-resumable-js

    @cached_property
    def request_data(self):
        return getattr(self.request, self.request.method)

    @cached_property
    def model_upload_field(self):
        content_type = ContentType.objects.get_for_id(self.request_data['content_type_id'])
        return content_type.model_class()._meta.get_field(self.request_data['field_name'])

    def post(self, request, *args, **kwargs):
        chunk = request.FILES.get('file')
        r = ResumableFile(self.model_upload_field, user=request.user, params=request.POST)
        if not r.chunk_exists:
            r.process_chunk(chunk)
        if r.is_complete:
            return HttpResponse(r.collect())
        return HttpResponse('chunk uploaded')

    def get(self, request, *args, **kwargs):
        r = ResumableFile(self.model_upload_field, user=request.user, params=request.GET)
        if not r.chunk_exists:
            return HttpResponse('chunk not found', status=204)
        if r.is_complete:
            return HttpResponse(r.collect())
        return HttpResponse('chunk exists')


admin_resumable = login_required(UploadView.as_view())



from .models import Foo

from django.test import client as client_module
from django.conf import settings
from django.contrib.contenttypes.models import ContentType

from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

import os
import pytest
import time


def create_test_file(file_path, size_in_megabytes):
    with open(file_path, "wb") as bigfile:
        bigfile.seek(size_in_megabytes * 1024 * 1024)
        bigfile.write(b"0")


def clear_uploads():
    upload_path = os.path.join(settings.MEDIA_ROOT, "admin_uploaded")
    if not os.path.exists(upload_path):
        return
    for the_file in os.listdir(upload_path):
        file_path = os.path.join(upload_path, the_file)
        try:
            if os.path.isfile(file_path):
                os.unlink(file_path)
        except Exception as e:
            print(e)


@pytest.mark.django_db
def test_fake_file_upload(admin_user, admin_client):
    foo_ct = ContentType.objects.get_for_model(Foo)
    clear_uploads()

    payload = client_module.FakePayload()

    def form_value_list(key, value):
        return [
            "--" + client_module.BOUNDARY,
            'Content-Disposition: form-data; name="%s"' % key,
            "",
            value,
        ]

    form_vals = []
    file_data = "foo bar foo bar."
    file_size = str(len(file_data))
    form_vals += form_value_list("resumableChunkNumber", "1")
    form_vals += form_value_list("resumableChunkSize", file_size)
    form_vals += form_value_list("resumableType", "text/plain")
    form_vals += form_value_list("resumableIdentifier", file_size + "-foobar")
    form_vals += form_value_list("resumableFilename", "foo.bar")
    form_vals += form_value_list("resumableTotalChunks", "1")
    form_vals += form_value_list("resumableTotalSize", file_size)
    form_vals += form_value_list("content_type_id", str(foo_ct.id))
    form_vals += form_value_list("field_name", "foo")
    payload.write(
        "\r\n".join(
            form_vals
            + [
                "--" + client_module.BOUNDARY,
                'Content-Disposition: form-data; name="file"; filename=foo.bar',
                "Content-Type: application/octet-stream",
                "",
                file_data,
                "--" + client_module.BOUNDARY + "--\r\n",
            ]
        )
    )

    r = {
        "CONTENT_LENGTH": len(payload),
        "CONTENT_TYPE": client_module.MULTIPART_CONTENT,
        "PATH_INFO": "/admin_resumable/upload/",
        "REQUEST_METHOD": "POST",
        "wsgi.input": payload,
    }
    response = admin_client.request(**r)
    assert response.status_code == 200
    upload_filename = file_size + "_foo.bar"
    upload_path = os.path.join(settings.MEDIA_ROOT, upload_filename)
    f = open(upload_path, "r")
    uploaded_contents = f.read()
    assert file_data == uploaded_contents


@pytest.mark.django_db
def test_fake_file_upload_incomplete_chunk(admin_user, admin_client):
    foo_ct = ContentType.objects.get_for_model(Foo)
    clear_uploads()

    payload = client_module.FakePayload()

    def form_value_list(key, value):
        return [
            "--" + client_module.BOUNDARY,
            'Content-Disposition: form-data; name="%s"' % key,
            "",
            value,
        ]

    form_vals = []
    file_data = "foo bar foo bar."
    file_size = str(len(file_data))
    form_vals += form_value_list("resumableChunkNumber", "1")
    form_vals += form_value_list("resumableChunkSize", "3")
    form_vals += form_value_list("resumableType", "text/plain")
    form_vals += form_value_list("resumableIdentifier", file_size + "-foobar")
    form_vals += form_value_list("resumableFilename", "foo.bar")
    form_vals += form_value_list("resumableTotalChunks", "6")
    form_vals += form_value_list("resumableTotalSize", file_size)
    form_vals += form_value_list("content_type_id", str(foo_ct.id))
    form_vals += form_value_list("field_name", "foo")
    payload.write(
        "\r\n".join(
            form_vals
            + [
                "--" + client_module.BOUNDARY,
                'Content-Disposition: form-data; name="file"; filename=foo.bar',
                "Content-Type: application/octet-stream",
                "",
                file_data[0:1],
                # missing final boundary to simulate failure
            ]
        )
    )

    r = {
        "CONTENT_LENGTH": len(payload),
        "CONTENT_TYPE": client_module.MULTIPART_CONTENT,
        "PATH_INFO": "/admin_resumable/admin_resumable/",
        "REQUEST_METHOD": "POST",
        "wsgi.input": payload,
    }
    try:
        admin_client.request(**r)
    except AttributeError:
        pass  # we're not worried that this would 500

    get_url = "/admin_resumable/admin_resumable/?"
    get_args = {
        "resumableChunkNumber": "1",
        "resumableChunkSize": "3",
        "resumableCurrentChunkSize": "3",
        "resumableTotalSize": file_size,
        "resumableType": "text/plain",
        "resumableIdentifier": file_size + "-foobar",
        "resumableFilename": "foo.bar",
        "resumableRelativePath": "foo.bar",
        "content_type_id": str(foo_ct.id),
        "field_name": "foo",
    }

    # we need a fresh client because client.request breaks things
    fresh_client = client_module.Client()
    fresh_client.login(username=admin_user.username, password="password")
    get_response = fresh_client.get(get_url, get_args)
    # should be a 404 because we uploaded an incomplete chunk
    assert get_response.status_code == 404


@pytest.mark.django_db
def test_real_file_upload(admin_user, live_server, driver):
    test_file_path = "/tmp/test_small_file.bin"
    create_test_file(test_file_path, 5)

    driver.get(live_server.url + "/admin/")
    driver.find_element(By.ID, "id_username").send_keys("admin")
    driver.find_element(By.ID, "id_password").send_keys("password")
    driver.find_element(By.XPATH, '//input[@value="Log in"]').click()
    driver.implicitly_wait(2)
    driver.get(live_server.url + "/admin/tests/foo/add/")
    WebDriverWait(driver, 10).until(EC.presence_of_element_located((By.ID, "id_bar")))
    driver.find_element(By.ID, "id_bar").send_keys("bat")
    driver.find_element(By.ID, "id_foo_input_file").send_keys(test_file_path)
    status_text = driver.find_element(By.ID, "id_foo_uploaded_status").text
    print("status_text", status_text)
    i = 0
    while i < 5:
        if "Uploaded" in driver.find_element(By.ID, "id_foo_uploaded_status").text:
            return  # success
        time.sleep(1)
        i += 1
    assert False, f"Status text is '{driver.find_element(By.ID, 'id_foo_uploaded_status').text}'; expected 'Uploaded"


@pytest.mark.django_db
def test_real_file_upload_with_upload_to(admin_user, live_server, driver):
    test_file_path = "/tmp/test_small_file.bin"
    create_test_file(test_file_path, 5)

    driver.get(live_server.url + "/admin/")
    driver.find_element(By.ID, "id_username").send_keys("admin")
    driver.find_element(By.ID, "id_password").send_keys("password")
    driver.find_element(By.XPATH, '//input[@value="Log in"]').click()
    driver.implicitly_wait(2)
    driver.get(live_server.url + "/admin/tests/foo/add/")
    WebDriverWait(driver, 10).until(EC.presence_of_element_located((By.ID, "id_bar")))
    driver.find_element(By.ID, "id_bar").send_keys("bat")
    driver.find_element(By.ID, "id_bat_input_file").send_keys(test_file_path)
    status_text = driver.find_element(By.ID, "id_bat_uploaded_status").text
    print("status_text", status_text)
    i = 0
    while i < 5:
        if "Uploaded" in driver.find_element(By.ID, "id_bat_uploaded_status").text:
            return  # success
        time.sleep(1)
        i += 1
    assert False, f"Status text is {driver.find_element(By.ID, 'id_bat_uploaded_status').text}; Expected 'Uploaded'"



import os
from setuptools import setup

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-async-upload-kg',
    version='4.0.5',
    packages=['admin_async_upload'],
    include_package_data=True,
    package_data={
        'admin_async_upload': [
            'templates/admin_resumable/admin_file_input.html',
            'templates/admin_resumable/user_file_input.html',
            'static/admin_resumable/js/resumable.js',
        ]
    },
    license='MIT License',
    description='A Django app for the uploading of large files from the django admin site.',
    long_description=README,
    url='https://github.com/bit/django-async-upload',
    author='Kajala Group Ltd.',
    author_email='support@kajala.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'Django>=3.0.14',
    ],
    tests_require=[
        'pytest-django',
    ]
)
